// OrderGenerator.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;

/**
 * The OrderGenerator is used at present to simulate the market behavior.
 *
 * @author Pietro Terna
 */
public class OrderGenerator  extends SwarmObjectImpl{

    /** How many unit we have, used at present to define the max value
	used to generate the 'digits' composing our recipes */

    /** the max length of a recipe */
    public int maxStepNumber;

    /** the max length of a step in a recipe */
    public int maxStepLength;

    /** used to record the number of generated orders*/
    public int orderCount;

    /** a specific order*/
    public Order anOrder;
    /** the list containig all the operating units*/
    public ListImpl unitList;
    /** the list containig all the orders*/
    public ListImpl orderList;
    /** the model (to which we apply mark up of the order revenue */
    VEFrameModelSwarm vEFrameModelSwarm;

    /** the integer vector containing a specific production recipe */
    public int[] orderRecipe1; 

    /** the integer vector containing the unit of time spent for each step 
     *  in the recipe; il will determine the tranformation of the orderRecipe1 
     *  vector into the orderRecipe2 vector containing the repetitions
     *  of the steps, to reproduce the length of each step in time
     *
     *  as an examples
     * 
     *  orderRecipe1      contains  1 12  7  3
     *  stepLengthInOrder contains  1  3  2  2
     *
     *  orderRecipe2      contains  1 12 12 12  7  7  3  3 
     */
    public int[] stepLengthInOrder;

    /** the integer vector containing a specific production recipe,
     *  developed following the length of each step
     */
    public int[] orderRecipe2; 

    /**
     * the constructor for OrderGenerator
     */
    public OrderGenerator (Zone aZone, int msn, int msl, ListImpl ul, 
			   ListImpl ol, VEFrameModelSwarm mo) 
    {

	// Call the constructor for the OrderGenerator parent class.
	super(aZone);

	orderCount=0;
	maxStepNumber=msn;
	maxStepLength=msl;
	unitList=ul;
        orderList=ol;

	orderRecipe1 = new int[maxStepNumber];
	stepLengthInOrder = new int[maxStepNumber];
	orderRecipe2 = new int[maxStepNumber*maxStepLength];

        vEFrameModelSwarm = mo;
    }

    /** creating an order form, choosing its recipe length and content */
    public void createRandomOrderWithNSteps() {
	int i, ii, iii, count, randomStepNumber;
	boolean operatingUnitNotFound;
	Unit aUnit;

	orderCount++;

	randomStepNumber=Globals.env.uniformIntRand.getIntegerWithMin$withMax
		    (1, maxStepNumber);

	// generating the sequence in the recipe

	// putting 0 in the vector is not strictly necessary, but may be
	// useful for future software modifications
	for (i=0;i<maxStepNumber;i++)orderRecipe1[i]=0;

	for (i=0;i<randomStepNumber;i++)
	    orderRecipe1[i]=
		Globals.env.uniformIntRand.getIntegerWithMin$withMax
		(1, unitList.getCount());

	// generating the length of each step

	// putting 0 in the vector is not strictly necessary, but may be
	// useful for future software modifications
	for (i=0;i<maxStepNumber;i++)stepLengthInOrder[i]=0;

	if (maxStepLength>1)
	    {
		for (i=0;i<randomStepNumber;i++)
		    stepLengthInOrder[i]=
			Globals.env.uniformIntRand.getIntegerWithMin$withMax
			(1, maxStepLength);
	    }
	else 
	    {
		for (i=0;i<randomStepNumber;i++)stepLengthInOrder[i]=1;
	    }

	// putting (i) sequence and (ii) length, together
	count=0;
	for (i=0;i<randomStepNumber;i++)
	    for (ii=0;ii<stepLengthInOrder[i];ii++)
		{
		    orderRecipe2[count]=orderRecipe1[i];
		    count++;
		}

	// creating an order
	anOrder = new Order(getZone(), orderCount,
			    Globals.env.getCurrentTime(),
			    randomStepNumber, orderRecipe2, vEFrameModelSwarm);

        // add the active orders to the general order list (they will be
	// eliminated when dropped in a unit, being finished); this
	// list has been introduced for accounting purposes [may be it would
	// be better subtitute it with a get to the units to know their
	// waiting lists] 
        orderList.addLast(anOrder);
	/**
	 * sending the order to the first production unit
	 * (we are acting as the Front End of the VE)
	 */
	operatingUnitNotFound=true;
	for (i=0;i<unitList.getCount() && operatingUnitNotFound;i++)
	    {
		aUnit=(Unit) unitList.atOffset(i);
		for (iii=0;iii<aUnit.getNOfPhasesToDealWith() &&  
			 operatingUnitNotFound;iii++)
		    {
		       if(anOrder.getNextStep()==aUnit.getProductionPhase(iii))
			    {
				operatingUnitNotFound=false;
				aUnit.setWaitingList(anOrder);
			    }
		    }
	    }
    }
}














