// Order.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;
/**
 * The form containing our Order metaphoric object
 *
 * @author Pietro Terna (with Paolo Pelligra and Dario Landini contibutions)
 */
public class Order  extends SwarmObjectImpl{

    /** the number of steps in the product recipe */
    public int stepNumber;
    /** the number of this order */
    public int orderNumber;
    /** the day in which the order as been generated */
    public int startingDay;
    /** integer vector, containing the recipe steps */
    public int[] orderRecipe;
    /** integer vector, recording the performed phases */
    public int[] orderState;
    /** the sequential number of the recipe referring to which
     * a news will be sent */
    public int numberOfTheStepForWhichANewsIsSent;
    /** the fixed and variable costs of the order */
    public float cumulatedFixedCost, cumulatedVariableCost;
    /** the sum of variable and fixed costs of the order */
    public float orderTotalCosts = 0;
    /** the model */
    VEFrameModelSwarm vEFrameModelSwarm;

  /**
   * constructor for Order
   */
  public Order (Zone aZone, int n, int sd, int sn, int [] r, VEFrameModelSwarm model) {

     // Call the constructor for the Order parent class
     super(aZone);

     orderNumber=n;
     stepNumber=sn;
     startingDay=sd;
     orderRecipe= new int[stepNumber];
     orderState = new int[stepNumber];
     // the model
     vEFrameModelSwarm = model;

     int i;

     for (i=0;i<stepNumber;i++)orderRecipe[i]=r[i];

     // announcing the order presence to the console
     if(StartVEFrame.verbose)
	 System.out.println("New order (#" + orderNumber + ")");

     // setting the state vector (0=empty; 1=done)
     if(StartVEFrame.verbose)System.out.print("orderState  vector ");
     for (i=1;i<=stepNumber;i++) {
	 orderState[i-1]=0;
	 if(StartVEFrame.verbose)System.out.print(orderState[i-1] + " ");
     }
     if(StartVEFrame.verbose)System.out.println();

     // announcing the recipe vector
     if(StartVEFrame.verbose)System.out.print("orderRecipe vector ");
     for (i=1;i<=stepNumber;i++)
	 if(StartVEFrame.verbose)System.out.print(orderRecipe[i-1] + " ");
     if(StartVEFrame.verbose)System.out.println();
  }

  /**
   * production next step in Order (it is in the orderRecipe vector,
   * at the same index value in which we find the first 0 in orderState
   * vector)
   */
    public int getNextStep () {

	int i, stepN;

	stepN=0;
	for (i=0;i< stepNumber && stepN==0;i++)
	    if (orderState[i]==0)stepN=orderRecipe[i];

	return stepN;
  }

  /**
   * Done step in Order
   */
    public void setDoneStep () {

	int i, done;

	done=0;
	for (i=0;i<= stepNumber && done==0;i++)
	    if (orderState[i]==0)
		{orderState[i]=1;
		done=1;
		}
    }

  /**
   * returning the orderNumber of this order
   */
    public int getOrderNumber () {
	return orderNumber;
  }

  /**
   * returning the stepNumber
   */
    public int getStepNumber () {

	return stepNumber;
  }

  /**
   * returning the startingDay
   */
    public int getStartingDay () {

	return startingDay;
  }

  /**
   * returning how many steps remain to terminate the production
   */
    public int getHowManyStepsToBeDone () {

	int i, stepsToBeDone;
	for (i=0;i<stepNumber && orderState[i]==1;i++); // to find the first position
	                                                // where orderState[i] != 1

	stepsToBeDone=stepNumber-i;
   
	if(StartVEFrame.verbose)
	    System.out.println
	       ("stepsToBeDone  " + stepsToBeDone+ " in order #"+ orderNumber);

	return stepsToBeDone;
  }

  /**
   * returning the production phase at a specific step
   */

    public int getProductionToBeDoneAtStepAheadNumber (int sn) {

	int i;
	int step;

	step = sn;
	for (i=0;i< stepNumber && orderState[i]==1;i++); // to find the first position
	                                                 // where orderState[i] != 1

	numberOfTheStepForWhichANewsIsSent=i+step-1;

	return orderRecipe[numberOfTheStepForWhichANewsIsSent];
    }

 /**
   * returning the sequential number of the recipe referring to which
   * a news will be sent
   */
    public int getNumberOfTheStepForWhichANewsIsSent () {

	return numberOfTheStepForWhichANewsIsSent;
  }

  /** returning the value of the production already done */
    public float getOrderEvaluation(){
	int globalOrderState = 0;
	float value = 0;

	// evaluate the number of steps done
	for(int i=0; i<orderState.length; i++){
	    globalOrderState += orderState[i];
	}

	value = globalOrderState * 
	    vEFrameModelSwarm.getRevenuePerEachRecipeStep();
	return value;
    }

  /** setting fixed and variable costs for the orders in production
   *  or finished */
    public void setFixedAndVariableOrderCosts(float fc, float vc){

	cumulatedFixedCost += fc;
	cumulatedVariableCost += vc;
	orderTotalCosts = (cumulatedFixedCost + cumulatedVariableCost);
  }

  /** getting the total of costs of the order */ 
    public float getOrderTotalCosts(){
	return orderTotalCosts;
    }

}
