// Unit.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;

/**
 * The Unit class instances are micro unit in our virtual enterprise;
 * i.e. the units where the steps required to build a product (to fulfill
 * an order) are done
 *
 * @author Pietro Terna (with Marco Remondino, Dario Landini, Nicola Ormezzano and
 * Paolo Pelligra contributions)
 */
public class Unit extends SwarmObjectImpl{

    /** the number identifying the unit*/
    public int unitNumber;
    /** choosing if we use or not the warehouses
	(to provide or not inventories of our products) */
    public boolean useWarehouses;
    /** choosing if we use or not newses about incoming productions*/
    public boolean useNewses;
    /** a flag about using pieces in warehouse */
    public boolean madeProduction;
    /** the list of all the units existing in our environment*/
    public ListImpl unitList;
    /** the list of the orders to be executed */
    public ListImpl  waitingList;
    /** the list containing the orders (if any) performed in a day, also
	via the inventories contained in our warehouse */
    public ListImpl dailyProductionList;
    /** the modelSwarm (to which we apply the setProductionTimeAndRecipeLength(...)
	method */
    VEFrameModelSwarm vEFrameModelSwarm;
    /** the temporary address of an existing unit */
    public Unit aUnit;
     /** the temporary address of a unit to wich send a news */
    public Unit receivingUnit;
    /** our warehouse, if any */
    public Warehouse myWarehouse;
    /** the first order to be executed */
    public Order firstOrder;
    /** an order inserted in a news */
    public Order incomingOrder;
    /** a news to be sent */
    public News aNews;
    /** a news to be deleted a the arrive of the related order */
    public News toBeDroppedNews;
    /** our rule master about inventories */
    public InventoryRuleMaster inventoryRuleMaster;
     /** our rule master about information system */
    public InformationRuleMaster informationRuleMaster;
    /** the list of the newses */
    public ListImpl newsList;
    /** how deep we look at production steps to propagatenews (how many steps 
     * beyond the current production phase)*/
    public int informationDeepness;
    /** of order for which newses are propagated */
    public int nOfOrdersInNewses;
    /** the whole list of the orders */
    public ListImpl wholeOrderList;
    /** our fixed and variable costs */
    public float fixedCosts, variableCosts;
    /** total daily variable costs */
    public float totalDailyVariableCosts;
    /** total daily fixed and variable costs */
    public float totalDailyCosts;
    /** the costs from the beginning of the simulation */
    public float totalCosts = 0;
    /** the daily revenues from the finished orders (sold) */
    public float saleRevenues;
    /** the revenues from the beginning of the simulation */
    public float totalSaleRevenues = 0;
    /** the costs of the finished orders */
    public float totalFinishedOrderCosts;
    /** the daily evaluation of the stored component */
    public float dailyStoredComponentValue = 0;
    /** the cumulative inventory financial costs generated from inventories */
    float totalInventoryFinancialCosts = 0;

    /**
     * the first Constructor for Unit (using warehouses)
     */
    public Unit (Zone aZone, int un, ListImpl ul, VEFrameModelSwarm mo,
		 InventoryRuleMaster rm, Warehouse wh, float fc, float vc, ListImpl ol)
    {
	// Call the constructor for the Unit parent class.
	super(aZone);

	useWarehouses = true;      // if we are using this constructor,
	                           // warehouses are activated
	useNewses = false;         // if we are using this constructor,
	                           // newses are not activated
        inventoryRuleMaster = rm;  // our InventoryRuleMaster
	myWarehouse = wh;	   // our warehouse address

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// the unit id number
	unitNumber=un;
	// the list of all the units
	unitList=ul;
	// the model
	vEFrameModelSwarm = mo;
	// the whole list of orders
        wholeOrderList = ol;
        // our fixed and variable costs
        variableCosts = vc;
        fixedCosts = fc;
	// the daily list of finished orders
	dailyProductionList= new ListImpl(getZone());

	// announce the unit to the console
	if(StartVEFrame.verbose)
	    System.out.println("Unit number " + unitNumber + " has been created.");
    }

    /**
     * the second Constructor for Unit (without warehouses)
     */
    public Unit (Zone aZone, int un, ListImpl ul, VEFrameModelSwarm mo,
                  float fc, float vc, ListImpl ol) {

	// Call the constructor for the Unit parent class
	super(aZone);

	useWarehouses = false;      // if we are using this constructor,
	                            // warehouses are not activated
	useNewses = false;          // if we are using this constructor,
	                            // newses are not activated
        inventoryRuleMaster = null; // to avoid compiler warning/error
	myWarehouse = null;         // " " "

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// the unit id number.
	unitNumber=un;
	// the list of all the units
	unitList=ul;
	// the model
	vEFrameModelSwarm = mo;
	// the daily list of finished orders
	dailyProductionList= new ListImpl(getZone());
	// the whole list of orders
        wholeOrderList = ol;
        // our fixed and variable costs
        variableCosts = vc;
        fixedCosts = fc;

	// announce the unit to the console
	if(StartVEFrame.verbose)
	    System.out.println("Unit number " + unitNumber + " has been created.");
    }

    /**
     * the third Constructor for Unit (with warehouses and newses)
     */
    public Unit (Zone aZone, int un, ListImpl ul, VEFrameModelSwarm mo,
		 InventoryRuleMaster rm, Warehouse wh, InformationRuleMaster irm, 
		 int idp, int noin, float fc, float vc, ListImpl ol) 
    {
	// Call the constructor for the Unit parent class.
	super(aZone);

	useWarehouses = true;       // if we are using this constructor,
	                            // warehouses are activated
        useNewses = true;           // if we are using this constructor,
	                            // newses are activated
	inventoryRuleMaster = rm;   // our InventoryRuleMaster
	informationRuleMaster= irm; // our InformationRuleMaster
        myWarehouse = wh;	    // our warehouse address

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// the unit id number.
	unitNumber=un;
	// the list of all the units
	unitList=ul;
	// the model
	vEFrameModelSwarm = mo;
	// the daily list if finished orders
	dailyProductionList= new ListImpl(getZone());
        // the news list
        newsList= new ListImpl(getZone());
        // how many steps we consider in recipes to propagate newses
        informationDeepness= idp;
	// # of orders for which newses are propagated
	nOfOrdersInNewses = noin;    
	// the whole list of orders
        wholeOrderList = ol;
        // our fixed and variable costs
        variableCosts = vc;
        fixedCosts = fc;

	// announce the unit to the console
	if(StartVEFrame.verbose)
	    System.out.println("Unit number " + unitNumber + " has been created.");
    }

    /** at the beginning of the 'day' each unit looks for the order(s) (if any)
     * to be executed; at present the production cycle is 'one order
     * a day, but we can also use inventories
     */
    public void unitStep1 ()
    {
        //add the beginnig of the day we have not yet generated any variable cost
        totalDailyVariableCosts = 0;
        totalDailyCosts = 0;

	if(StartVEFrame.verbose)
	    System.out.println("Step 1 in unit " + unitNumber +
			       "; waitingList count " + waitingList.getCount());

	// after the specific step in recipe production (directly or from 
	// inventories) we place the order(s) in a daily production
	// list, which will be used at unitStep2 to diffuse production
	// results

	// the production
        madeProduction=false;
	if(waitingList.getCount() > 0){
	      firstOrder=(Order)waitingList.removeFirst();
	      dailyProductionList.addLast(firstOrder);
                if (useWarehouses && ! myWarehouse.empty()){
                    myWarehouse.decreaseInventoryCounterValue();
                }
		else {
		    madeProduction=true;
		    addDailyVariableCosts();
		}
              firstOrder.setDoneStep();
              //set the costs of the order in production
              firstOrder.setFixedAndVariableOrderCosts(fixedCosts, variableCosts);

        }

	// having made no production, check to increase inventories
	if(! madeProduction && useWarehouses){
	    if(useNewses){
		if(inventoryRuleMaster.increaseInventoriesUsingNewses
		   (unitNumber, myWarehouse, newsList)){
		    myWarehouse.increaseInventoryCounterValue();
		    //recording variable costs
		    addDailyVariableCosts();
                }
	    }
	    else {
                if(inventoryRuleMaster.increaseInventories
                                          (unitNumber, myWarehouse)){
		    myWarehouse.increaseInventoryCounterValue();
		    //recording variable costs
		    addDailyVariableCosts();
                }
	    }
        }

	// the use of inventories
	if (useWarehouses)
	    {
		while(waitingList.getCount()>0 && ! myWarehouse.empty())
		    {
		    myWarehouse.decreaseInventoryCounterValue();
		    firstOrder=(Order)waitingList.removeFirst();
		    firstOrder.setDoneStep();
		      // set the production costs into the order
		      firstOrder.setFixedAndVariableOrderCosts(
							   fixedCosts, variableCosts);
		    // is the next recipe step to be done in our unit? If yes, 
	            // assign the order immediately to ourselves (if there is 
	            // room in the inventories we can make the production step 
		    // in this same cycle)
		     if(getProductionPhase()==firstOrder.getNextStep())
			 waitingList.addLast(firstOrder);
		    // if not, put the order in the daily production list
		     else dailyProductionList.addLast(firstOrder);
		    }
	    }
    }

    /**
     * sending the orders to the subsequent production units
     */
    public void unitStep2 () {

	int i;
        int x;
	boolean operatingUnitNotFound;
	saleRevenues = 0;
        totalFinishedOrderCosts = 0;

	if(StartVEFrame.verbose)
	    System.out.println("Step 2 in unit " + unitNumber +
			       " dailyProductionList count " +
			       dailyProductionList.getCount());

	while(dailyProductionList.getCount()>0)
	    {
		firstOrder=(Order)dailyProductionList.removeFirst();

		// do we have concluded the production step?
		// if yes ... we drop the object
		if (firstOrder.getNextStep()==0)
		    {
			if(StartVEFrame.verbose)
			    System.out.println("drop order # " +
					       firstOrder.getOrderNumber() +
					       " startingDay " +
					       firstOrder.getStartingDay() +
					       " endDay " +
					       Globals.env.getCurrentTime() +
					       " stepNumber " +
					       firstOrder.getStepNumber() +
					       " ratio " +
					       ( (float)
						(Globals.env.getCurrentTime() -
						firstOrder.getStartingDay()))/
					       firstOrder.getStepNumber());

			// the production is concluded, so we inform
			// our environment (the model) of the time spent
			// and of the # of steps in recipe
			vEFrameModelSwarm.setProductionTimeAndRecipeLength(
			  Globals.env.getCurrentTime() -
			  firstOrder.getStartingDay(),
			  firstOrder.getStepNumber());
			if(StartVEFrame.verbose)
			    System.out.println("Order #" +
					       firstOrder.orderNumber +
					       " has been completed");
			// remove the dropped order from the wholeOrderList
			wholeOrderList.remove(firstOrder);
                        // recording the revenue from the finished order
                        saleRevenues += firstOrder.getOrderEvaluation();
                        //recording the costs of the order
                        totalFinishedOrderCosts += firstOrder.getOrderTotalCosts();
                        // now we drop the object (also to save memory space)
                        firstOrder.drop();
		    }

		// if not ... we send it to the subsequent production unit
		else
		    {
			operatingUnitNotFound=true;
			for (i=0;i<unitList.getCount() &&
				 operatingUnitNotFound;i++)
			    {
				aUnit=(Unit)unitList.atOffset(i);

				if (firstOrder.getNextStep() ==
				    aUnit.getProductionPhase())
				    {
					operatingUnitNotFound=false;
					aUnit.setWaitingList(firstOrder);
				    }
			    }
		    }
	    }

	// set accounting data
	setTotalDailyCosts();
	setTotalCosts();
	setTotalSaleRevenues();
	if(useWarehouses){
	    setTotalInventoryFinancialCosts();
	    setStoredComponentDailyValue();
	}

    }

    public void unitStep3 () {

	int n;
	int i;
	int m;
        int stepsToBeDone;
        int receivingUnitNumber;
	boolean operatingUnitNotFound;

	for(n=0; n<nOfOrdersInNewses && waitingList.getCount()>n && useNewses; n++) {
		// we get the first order from the waitingList to find how many steps 
		// are necessary to finish its recipe
		incomingOrder=(Order)waitingList.atOffset(n);
		stepsToBeDone=incomingOrder.getHowManyStepsToBeDone();

		// we ask the order about the subsequent production units, according 
		// information deepness (we start from next step, in position 2)
		for (i=2;i<=stepsToBeDone && i<=informationDeepness+1;i++)
		    {
			receivingUnitNumber = 
			    incomingOrder.getProductionToBeDoneAtStepAheadNumber(i);
			      if(StartVEFrame.verbose)
			          System.out.println(
                                        " Receiving Unit # " +
				        incomingOrder.
				        getProductionToBeDoneAtStepAheadNumber(i));
			if(informationRuleMaster.
			   checkToSendNews(unitNumber, receivingUnitNumber)
			   && unitNumber!= receivingUnitNumber)
			    {
				operatingUnitNotFound=true;
				// searching the units able to make the next production
				// steps and sending the newses
				for (m=0;
				     m<unitList.getCount() && operatingUnitNotFound;
				     m++)
				    {
					receivingUnit=(Unit)unitList.atOffset(m);
					if (receivingUnitNumber==receivingUnit.
					    getProductionPhase())
					    {
						aNews=new News (getZone(), unitNumber, 
								receivingUnit.
								getProductionPhase(), 
								incomingOrder);
						receivingUnit.setNewsList(aNews);
						operatingUnitNotFound=false;
					    }
				    }
			    }
		    }
	    }
    }
    /**
     * return the production phase of the unit
     */
    public int getProductionPhase()
    {
	return unitNumber; // at present unitNumber and production phase are
	                   // the same; this is a tempory solution
    }

    /**
     * Putting an order in the production waitingList of the unit
     * and deleting the related news
     */
    public void setWaitingList(Order anOrder)
    {
       boolean newsToBeDroppedNotFound;
       int i; 

        waitingList.addLast(anOrder);
	if(StartVEFrame.verbose)
	    System.out.println("The unit " + unitNumber +
			       " has received the order # "
			       + anOrder.getOrderNumber());

	if (useNewses && newsList.getCount()>0)
	    //looking for a news containing information about
	    // the order just arrived to the unit
	    for (i=1;i<=newsList.getCount();i++)
		{
		    toBeDroppedNews=(News)newsList.removeFirst();
		    if (toBeDroppedNews.incomingOrderInNews.getOrderNumber()
			==anOrder.getOrderNumber())
			{
			    toBeDroppedNews.drop();
			    if(StartVEFrame.verbose)
				System.out.println(
					"Dropping the news announcinig the order n "
					+ toBeDroppedNews.incomingOrderInNews.
					                     getOrderNumber());
			}
		    else
			newsList.addLast(toBeDroppedNews);
		}
    }

    /**
     * return the waiting list length
     */
    public int getWaitingListLength()
    {
	return waitingList.getCount();
    }

    /**
     * putting a news in the newsList of the unit after
     * checking list to avoid duplicated news
     */
    public void setNewsList(News aNews)
    {
	int i;
	boolean  duplicatedNews;
	
	// a single news of the newsList
	News controlNews;
	// the news just arrived
	News toBeCheckedNews;

	toBeCheckedNews=aNews;
	duplicatedNews=false;

	if (newsList.getCount()>0)
	    {
		for (i=0;i<newsList.getCount() && ! duplicatedNews;i++)
		    {
			controlNews=(News) newsList.removeFirst();
			newsList.addLast(controlNews);
			if (toBeCheckedNews.incomingOrderInNews.getOrderNumber()==
			    controlNews.incomingOrderInNews.getOrderNumber()
			    && toBeCheckedNews.getNumberOfTheStepForWhichANewsIsSent()==
			    controlNews.getNumberOfTheStepForWhichANewsIsSent())

			    duplicatedNews=true;
		    }

		if (! duplicatedNews)
		    {
                  newsList.addLast(toBeCheckedNews);
                  if(StartVEFrame.verbose)
		      System.out.println("Unit " + unitNumber +
					 " accept a news containing order # "
					 + toBeCheckedNews.incomingOrderInNews.
					 getOrderNumber());
		    }
		else
		    if(StartVEFrame.verbose)
			System.out.println("Unit " + unitNumber +
					   " doesn't accept a news containing order # "
					   + toBeCheckedNews.incomingOrderInNews.
					   getOrderNumber()+
					   " which just exists for the same step");
            }
	else
	    {
		newsList.addLast(toBeCheckedNews);
		if(StartVEFrame.verbose)
		    System.out.println("Unit " + unitNumber +
				       " accepts a news containing order N "
				       + toBeCheckedNews.incomingOrderInNews.
				       getOrderNumber()+
				       " because newsList is empty");
	    }
    }

    /** set the value of the total daily variable costs */
    public void addDailyVariableCosts(){
        totalDailyVariableCosts += variableCosts;
    }

    /** set the global sum of fixed and variable costs */
    public void setTotalDailyCosts(){
      totalDailyCosts = fixedCosts + totalDailyVariableCosts;
    }

    /** cumulate the total daily costs */
    public void setTotalCosts(){
      totalCosts += totalDailyCosts;
    }

    /** return total daily costs */
    public float getTotalDailyCosts(){
      return totalDailyCosts;
    }

    /** return total costs */
    public float getTotalCosts(){
      return totalCosts;
    }

    /** record the revenue from a dropped order */
    public float getSaleRevenues(){
      return saleRevenues;
    }

    /** cumulatere the revenues from finished orders */
    public void setTotalSaleRevenues(){
      totalSaleRevenues += saleRevenues;
    }

    /** return the total revenue */
    public float getTotalSaleRevenues(){
      return totalSaleRevenues;
    }

    /** return total costs from finished orders */
    public float getTotalFinishedOrderCosts(){
	return totalFinishedOrderCosts;
    }

    /** record the cumulated financial costs of inventories */
    public void setTotalInventoryFinancialCosts(){
        totalInventoryFinancialCosts += (myWarehouse.getInventoryQuantity()*
             (vEFrameModelSwarm.getInventoryFinancialRate()/200));
	        // getInventoryFinancialRate returns an yearly rate, which is here
	        // divided by 200 day of labour per year in an approximate way
    }

    /** return the financial costs of inventories */
    public float getTotalInventoryFinancialCosts(){
        return totalInventoryFinancialCosts;
    }

    /** record the daily value of inventories */
    public void setStoredComponentDailyValue(){
        if(vEFrameModelSwarm.getInventoryEvaluationCriterion()==3)
	    dailyStoredComponentValue = myWarehouse.getInventoryQuantity()*
		vEFrameModelSwarm.getRevenuePerEachRecipeStep();

        else if(vEFrameModelSwarm.getInventoryEvaluationCriterion()==2)
	    dailyStoredComponentValue = myWarehouse.getInventoryQuantity()*
		(fixedCosts + variableCosts);

        else if(vEFrameModelSwarm.getInventoryEvaluationCriterion()==1)
	    dailyStoredComponentValue = myWarehouse.getInventoryQuantity()*
		variableCosts;

        else if(vEFrameModelSwarm.getInventoryEvaluationCriterion()!=1 &&
		vEFrameModelSwarm.getInventoryEvaluationCriterion()!=2 &&
		vEFrameModelSwarm.getInventoryEvaluationCriterion()!=3)
	    {
		System.out.println("your criterion doesn't exist: the variable"+
				   " inventoryEvaluationCriterion must 1, 2 or 3");
		System.exit(0);
	    }
    }

    public float getDailyStoredComponentValue(){
        return dailyStoredComponentValue;
    }

}



