//AccountingParameters.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;
import java.io.*;
import java.util.*;
/**
 * This class reads the variable and fixed cost parameters for each unit
 *
 * @author Dario Landini and Nicola Ormezzano; IO improved by Michele Sonnessa
 */

public class AccountingParameters extends SwarmObjectImpl {

  /** the vector of the fixed costs (per day or time step) of each unit */
  public float[] fixedCosts;
  /** the vector of the variable costs (per day or time step) of each unit */
  public float[] variableCosts;
  /** the number of the units (the size of the vectors) */
  public int totalUnitNumber;

  int i;

  /** the constructor of AccountingParameters
   */
  public AccountingParameters(Zone aZone, int un) {

      // call the constructor for the AccountingParameters parent class
      super(aZone);
      totalUnitNumber = un;
  }

  /** read the fixed cost parameters from the fixedCosts.txt file*/
  public void readFixedCosts(){
      try{
	  fixedCosts = new float[totalUnitNumber];
	  // read the file
	  BufferedReader fileInputFixedCosts = null;

	  fileInputFixedCosts = new 
	      BufferedReader(new FileReader("CostParameters/fixedCosts.txt"));

	  StringTokenizer t;

	  for(i = 0; i < fixedCosts.length; i++)
	      {
		  String line = fileInputFixedCosts.readLine();
		  t = new StringTokenizer(line, " ");
		  fixedCosts[i] = Float.parseFloat(t.nextToken());
	      }

      }
 
      catch (FileNotFoundException fnfe)
	  {
	      System.out.println(fnfe);
	      System.exit(0);
	  } 
      catch (IOException e)
	  {
	      System.out.println("IOException:" + e.toString());
	  }
  }

  /** read the variable cost parameters from the variableCosts.txt file */
  public void readVariableCosts(){
      try{
	  variableCosts = new float[totalUnitNumber];
	  // read the file
	  BufferedReader fileInputVariableCosts = null;
	  
	  fileInputVariableCosts = new 
	      BufferedReader(new 
		  FileReader("CostParameters/variableCosts.txt"));

	  StringTokenizer t;

	  for(i=0; i<variableCosts.length; i++)
	      {
		  String line = fileInputVariableCosts.readLine();
		  t = new StringTokenizer(line, " ");
		  variableCosts[i] = Float.parseFloat(t.nextToken());
	      }
      }
      catch (FileNotFoundException fnfe)
	  {
	      System.out.println(fnfe);
	      System.exit(0);
	  }

      catch (IOException e)
	  {
	      System.out.println("IOException:" + e.toString());
	  }
  }

    /** returning the fixed cost for a specific unit */
    public float getFixedCosts(int i){
	return fixedCosts[i-1];
    }

    /** returning the variable cost for a specific unit */
    public float getVariableCosts(int i){
	return variableCosts[i-1];
    }

}

